"""
常用批量图片增强函数
"""
from ..blending import ObjectReplaceBlend, DirectBlending, PyramidBlending, PoissonBlending
from ..annonation import Text2XML
import logging
import os
from tqdm import tqdm
from PIL import Image
from xl_tool.xl_io import file_scanning
from xl_tool.data.image.annonation import get_bndbox
from xl_tool.xl_io import save_to_json, read_json
from random import choice
import numpy as np


def update_bg_config_file(background_config_path):
    """
    背景配置文件，voc标注（标注用于合成的区域）格式文件和图片放置同一文件夹即可
    """
    files = file_scanning(background_config_path, "xml", sub_scan=False)
    configs = []
    for file in files:
        temp = {}
        temp["file"] = file.replace("xml", "jpg")
        box = get_bndbox(file)[0]
        temp["region"] = [int(i) for i in box["coordinates"]]
        configs.append(temp)
    save_to_json(configs, f"{background_config_path}/background_config.json", indent=4)


def batch_object_replace(labeled_data, object_files, object_classes, image_save_path, xml_save_path=None,
                         aspect_jump=0.5, aspects=None, replace_classes=None):
    """
    批量替换数据增强
    Args:
        labeled_datas: [(image_file,xml_file), ...]
        object_files: 目标框文件列表
        object_classes： 目标类别列表，应该与目标框文件列表长度一致
        image_save_path： 增强图片保存路径
        xml_save_path： xml保存路径
        aspect_jump: 是否对长宽比进行扰动，最终用于匹配的长宽比为以下范围采样：
                原始长宽比-aspect_jump ， 原始长宽比+aspect_jump
        aspects: 目标框长宽比，None,则会自动读取图片生成
        replace_classes: 替换的类别列表，None表示替换所有类别
    """
    blender = ObjectReplaceBlend()
    object_images = [Image.open(i) for i in object_files]
    aspects = [i.size[0] / i.size[1] for i in object_images] if not aspects else aspects
    xml_save_path = xml_save_path if xml_save_path else image_save_path
    pbar = tqdm(list(labeled_data))

    assert len(aspects) == len(object_images), "目标长宽比数量与图片数量不一致"
    assert len(object_classes) == len(object_images), "目标类别数量与图片数量不一致"

    object_images, aspects = zip(*sorted(zip(object_images, aspects), key=lambda x: x[1]))
    for image_file, xml_file in pbar:
        try:
            # assert os.path.basename(xml_file).split(".")[0] == os.path.basename(image_file).split(".")[0], "图片与标注文件无法对应"
            xml_folder = r"Dataset"
            xml_source = r'Dataset'
            aug_image, boxes, aug_object_indexes = blender.blending_one_image(image_file, object_images, aspects,
                                                                              xml_file,
                                                                              random_choice=False,
                                                                              aspect_jump=aspect_jump,
                                                                              replace_classes=replace_classes)
            save_img = f"{image_save_path}/{'replace_aug_' + os.path.basename(image_file)}"
            aug_image.save(save_img)
            text2xml = Text2XML()
            filename = os.path.basename(xml_file)
            save_xml = f"{xml_save_path}/{'replace_aug_' + os.path.basename(xml_file)}"
            objects_info = [[object_classes[index]] + coordinate for coordinate, index in
                            zip(boxes, aug_object_indexes)]
            xml = text2xml.get_xml(xml_folder, filename, filename, xml_source, aug_image.size, objects_info)
            with open(save_xml, "w", encoding="utf-8") as f:
                f.write(xml)
        except Exception as e:
            logging.warning("数据替换异常！！！！\n" + str(e) + "\n" + str(xml_file))
        pbar.set_description("替换增强进度：")


def batch_mul_object_blend(object_path, background_config_path, image_save_path, xml_save_path=None,
                           blending_method="direct", categories=None, target_number=1000, num_per_images=(1, 2, 3, 4),
                           num_distribute=(0.1, 0.35, 0.4, 0.15), space_distribute=(0.5, 0.5),
                           folder="Dataset",
                           source='Dataset', ):
    """
    Args:
        object_path： 目标框位置
        background_config_path: 背景配置文件地址
        image_save_path: 图片保存位置
        xml_save_path: xml 文件保存位置
        categories: 用于合成的类别
        space_distribute: 左右排列与上下排列的比例

    """
    # todo 目前多目标只支持直接融合，不支持泊松融合和金字塔融合
    assert sum(num_distribute) == 1.0, "概率分布不为1"
    assert sum(space_distribute) == 1.0, "概率分布不为1"
    blender = DirectBlending()
    update_bg_config_file(background_config_path)
    background_config_file = f"{background_config_path}/background_config.json"
    xml_save_path = xml_save_path if xml_save_path else image_save_path
    os.makedirs(xml_save_path, exist_ok=True)
    os.makedirs(image_save_path, exist_ok=True)
    categories = os.listdir(object_path) if not categories else categories
    assert len(categories) >= 1, f"目标文件夹下未找到类别:{object_path}"

    # print(background_config_file)
    background_configs = read_json(background_config_file)
    logging.info(f"有效背景配置文件数量：{len(background_configs)}")
    cats_files = {cat: file_scanning(f"{object_path}/{cat}", file_format="jpg|JPG|jpeg|JPEG") for cat in categories}
    cats_files = {k: v for k, v in cats_files.items() if len(v) > 0}
    categories = list(cats_files.keys())
    logging.info("有效融合类别与数量分布：\n" + str({k: len(v) for k, v in cats_files.items()}))
    count = 0
    cats_count = {cat: 0 for cat in categories}
    num_state = 0
    number = num_per_images[num_state]
    sp_dis = 1 if number == 1 else np.random.choice([1, 2], space_distribute)
    choose_files = []
    while count < target_number:
        # 目前只支持并排排列，完全随机暂未实现（涉及比例和空间规划组合）
        try:
            ratio = count / target_number
            if ratio <= sum(num_distribute[:num_state + 1]):
                pass
            else:
                num_state += 1
                number = num_per_images[num_state]
                sp_dis = 1 if number == 1 else np.random.choice([1, 2], p=space_distribute)
            # 随机选择用于合成的类别，均布

            choose_cats = [choice(categories) for _ in range(number)]
            choose_files = [choice(cats_files[cat]) for cat in choose_cats]
            background_config = choice(background_configs)
            background_file, region = background_config["file"], background_config['region']
            image_file_name = f"aug_{number}_{count}.jpg"
            save_file = f"{image_save_path}/{image_file_name}"

            background_img, image_sizes, positions = blender.blending_images(background_file, choose_files,
                                                                             sp_dis,
                                                                             save_file,
                                                                             region)

            objects_info = [[cat, ] + list(positions[i]) for i, cat in enumerate(choose_cats)]
            text2xml = Text2XML()
            xml = text2xml.get_xml(folder, image_file_name, image_file_name, source, background_img.size, objects_info)
            xml_save_file = f"{xml_save_path}/{image_file_name.replace('jpg', 'xml')}"
            with open(xml_save_file, "w", encoding="utf-8") as f:
                f.write(xml)
            for c in choose_cats:
                cats_count[c] += 1
            count += 1
            logging.info("\r合成数量：" + str(count))
        except Exception as e:
            logging.warning("合成异常！！！：\n" + str(e) + str(choose_files))



#!usr/bin/env python3
# -*- coding: UTF-8 -*-
import os
import xml
from abc import abstractmethod
from .annonation import coordinate_name
from xl_tool.xl_io import file_scanning, time_analysis_wrapper, save_to_json
from xl_tool.xl_concurrence import MyThread
from math import ceil
from tqdm import tqdm
from copy import deepcopy
from collections import Counter
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator, MultipleLocator, FuncFormatter
plt.rcParams['font.sans-serif'] = ['SimHei']  # 步骤一（替换sans-serif字体）
plt.rcParams['axes.unicode_minus'] = False  # 步骤二（解决坐标轴负数的负号显示问题）


class ObjectAnalysis:
    def __init__(self):
        pass

    @staticmethod
    def get_aspect_ratio(x1, y1, x2, y2, w=None, h=None, ignore_zero=True):
        """
        get aspect ratio(w/h)
        """
        try:
            w = w if w else x2 - x1
            h = h if h else y2 - y1
            # assert w > 0, "width can not be zero"
            # assert h > 0, "height can not be zero"
            aspect_ratio = w / h
        except ZeroDivisionError:
            if ignore_zero:
                return 0
            else:
                raise ZeroDivisionError
        return aspect_ratio

    @staticmethod
    def get_box_info(box_coordinate: tuple, image_w, image_h, image_area=None, coordinate_format="xxyy",
                     ignore_zero=True):
        """get box width, height, area ratio relative to whole image
        Args:
            box_coordinate: tuple,(x1,y1,x2,y2) whene coordinate_format is xxyy else (x,y,w,h)
            image_w: int or float
            image_h: int or float
            image_area: int or float or None, default None
            coordinate_format: string ,default xxyy, if not equal xxyy, meaning x,y,w,h
            ignore_zero: bool, whether to ignore zero divide, if True, return zero when zero divide happen
        """
        if coordinate_format == "xxyy":
            x, y = box_coordinate[:2]
            w, h = box_coordinate[2] - box_coordinate[0], box_coordinate[3] - box_coordinate[1]
        else:
            x, y, w, h = box_coordinate
        try:
            # w_ratio = w / image_w
            # h_ratio = h / image_h
            image_area = image_area if image_area else (image_h * image_w)
            box_area = w * h
            area_ratio = box_area / image_area
            box_aspect = w / h
            return w, h, box_area, area_ratio, box_aspect
        except ZeroDivisionError:
            if ignore_zero:
                return 0, 0, 0, 0, 0
            else:
                raise ZeroDivisionError

    @abstractmethod
    def dataset_analysis(self, annotation_path, number_thread=2):
        pass


class VocAnalysis(ObjectAnalysis):
    """VOC dataset analysis"""

    def __init__(self):
        super(VocAnalysis, self).__init__()
        self.results_template = {
            "images": {"aspects": [], "areas": [], "box_per_images": []},
            "boxes": {"aspects": [], "areas": [], "labels": [], "ratios": []}
        }
        self.results = None

    def extract_annotation(self, xml_file):
        dom = xml.dom.minidom.parse(xml_file)
        doc = dom.documentElement
        bndboxes = []
        for object_node in doc.getElementsByTagName('object'):
            name = object_node.getElementsByTagName('name')[0].firstChild.data
            coordinates = [int(object_node.getElementsByTagName(i)[0].firstChild.data) for i in coordinate_name]
            bndboxes.append({"name": name, "coordinates": coordinates})
        temp_node = doc.getElementsByTagName('size')[0]
        w = float(temp_node.getElementsByTagName("width")[0].firstChild.data)
        h = float(temp_node.getElementsByTagName("height")[0].firstChild.data)
        return w, h, bndboxes

    def single_annotation_analysis(self, xml_file):
        w, h, boxes = self.extract_annotation(xml_file)
        image_area = w * h
        image_aspect = w / h
        box_per_image = len(boxes)
        _, _, box_areas, area_ratios, box_aspects = zip(
            *[self.get_box_info(box["coordinates"], None, None, image_area) for box in boxes])
        labels = [i["name"] for i in boxes]
        return {"image": (image_aspect, image_area, box_per_image),
                "box": (labels, box_aspects, box_areas, area_ratios)}




    def batch_annotation_analysis(self, xml_files):
        batch_results = {
            "images": {"aspects": [], "areas": [], "box_per_images": []},
            "boxes": {"aspects": [], "areas": [], "labels": [], "ratios": []}
        }
        pbar = tqdm(xml_files)
        for xml_file in pbar:
            try:
                results = self.single_annotation_analysis(xml_file)
            except ValueError:
                print("NonAnnotation", xml_file)
                continue
            except ZeroDivisionError:
                print("ZeroDivisionError", xml_file)
                continue
            batch_results["images"]["aspects"].append(results["image"][0])
            batch_results["images"]["areas"].append(results["image"][1])
            batch_results["images"]["box_per_images"].append(results["image"][2])
            # box
            batch_results["boxes"]["aspects"].extend(results["box"][1])
            batch_results["boxes"]["areas"].extend(results["box"][2])
            batch_results["boxes"]["labels"].extend(results["box"][0])
            batch_results["boxes"]["ratios"].extend(results["box"][3])
            # pbar.set_description("analysis progress: ")
        return batch_results

    @time_analysis_wrapper("分析耗时：")
    def dataset_analysis(self, annotation_path, number_thread=1, json_path=None, plt_path=None):
        """
        Args:
            annotation_path: 标注文件位置
            number_thread: 线程数量
            json_path : 用于保存分析结果文件名
            plt_path ：用于保存分析结果图片的路径
        Returns:
            Dict
        # Todo 图片结果未作分析，意义不大
        """
        self.results = deepcopy(self.results_template)
        xml_files = file_scanning(annotation_path, "xml", sub_scan=True)
        if number_thread == 1:
            self.results = self.batch_annotation_analysis(xml_files)
        else:
            threads = list(range(number_thread))
            handle_num = ceil(len(xml_files) / number_thread)
            for i in range(number_thread):
                args = (xml_files[i * handle_num:(i + 1) * handle_num],)
                threads[i] = MyThread(target=self.batch_annotation_analysis,
                                      args=args)
            for thread in threads:
                thread.start()
            for thread in threads:
                thread.join()
            for thread in threads:
                results = thread.get_result()
                self.results["images"]["aspects"].extend(results["images"]["aspects"])
                self.results["images"]["areas"].extend(results["images"]["areas"])
                self.results["images"]["box_per_images"].extend(results["images"]["box_per_images"])
                # box
                self.results["boxes"]["aspects"].extend(results["boxes"]["aspects"])
                self.results["boxes"]["areas"].extend(results["boxes"]["areas"])
                self.results["boxes"]["labels"].extend(results["boxes"]["labels"])
                self.results["boxes"]["ratios"].extend(results["boxes"]["ratios"])
        self.results["boxes"]["areas"] = [i // 100 for i in self.results["boxes"]["areas"]]
        self.results["images"]["areas"] = [i // 100 for i in self.results["images"]["areas"]]

        labels = Counter(self.results["boxes"]["labels"])
        if plt_path:
            fig = plt.figure(figsize=(min(64, len(labels) + 4), 12))
            ax = fig.add_subplot(1, 1, 1)
            ax.bar(*zip(*labels.items()))

            # plt.bar(*zip(*labels.items()))
            ax.set_xlabel("label")
            ax.set_ylabel("number")
            plt.xticks(rotation=45)
            # plt.show()
            plt.savefig(os.path.join(plt_path, "labels.png"))
        plt.cla()
        plt.figure(figsize=(16, 10))
        box_aspects = plt.hist(self.results["boxes"]["aspects"], bins=20, range=(0.01, 5))
        box_aspects = list(box_aspects[0]), list(box_aspects[1][1:])
        if plt_path:
            ax = plt.gca()
            ax.xaxis.set_minor_locator(AutoMinorLocator(5))
            ax.xaxis.set_major_locator(MultipleLocator(0.500))
            plt.xlabel("box_aspects")
            plt.ylabel("number")
            plt.savefig(os.path.join(plt_path, "box_aspects.png"))
        plt.cla()
        box_areas = plt.hist(self.results["boxes"]["areas"], bins=20)
        box_areas = list(box_areas[0]), list(box_areas[1][1:])
        if plt_path:
            plt.xlabel("box_areas")
            plt.ylabel("number")
            ax = plt.gca()
            ax.xaxis.set_minor_locator(AutoMinorLocator())
            plt.savefig(os.path.join(plt_path, "box_areas.png"))
        plt.cla()
        box_ratios = plt.hist(self.results["boxes"]["ratios"], bins=20)
        box_ratios = list(box_ratios[0]), list(box_ratios[1][1:])
        if plt_path:
            plt.xlabel("box_areas_ratio")
            plt.ylabel("number")
            ax = plt.gca()
            ax.xaxis.set_major_locator(MultipleLocator(0.0500))
            ax.xaxis.set_minor_locator(AutoMinorLocator())
            plt.savefig(os.path.join(plt_path, "box_areas_ratio.png"))

        analysis_result = {"labels": labels, "box_aspects": box_aspects,
                           "box_areas": box_areas, "box_ratios": box_ratios}
        if json_path:
            save_to_json(analysis_result, json_path)
        return analysis_result



#!usr/bin/env python3
# -*- coding: UTF-8 -*-
from .blending import PyramidBlending,PoissonBlending,ObjectReplaceBlend,DirectBlending,SegBlend
from .general import resize_with_pad,read_with_rgb,grey_world,cv_show_image,affine_with_rotate_scale,linear_contrast
from .config import IMAGE_FORMAT
from .annonation import Text2XML,get_bndbox,get_boximgs,Coco2Voc,xml_object_extract